from random import randint
import json
from typing import Optional, List

from .theme import Theme, DefaultTheme


class Game:
    def __init__(self, field_size: int = 5, theme: Theme = DefaultTheme, score_key: Optional[str] = None):
        self.snake = [(field_size // 2 + 1, field_size // 2 + 1)]  # field starts at 1; (1, 1) is the top left corner
        self.field_size = field_size
        self.orientation = 0  # 0 = Up, 1 = Right, 2 = Down, 3 = Left
        self.apples = []

        self.theme = theme

        self.max_apples = 1
        self.max_score = self.field_size ** 2
        self.orientation_switch = {  # urdl
            0: (0, -1),
            1: (1, 0),
            2: (0, 1),
            3: (-1, 0),
        }

        self.score_key = score_key if score_key else str(self.field_size)

        try:
            with open("scores.json") as f:
                scores = json.loads(f.read())
                self.scores = scores

                self.highscore = scores.get(self.score_key, 0)
        except (json.decoder.JSONDecodeError, FileNotFoundError) as e:
            self.highscore = 0
            self.scores = {}
        # ┏
        # ━
        # ┃
        # ┛
        # ┗
        # ┓

    def set_apple(self):
        if self.max_score > len(self.apples) + self.score:
            while apple_location := (randint(1, self.field_size), randint(1, self.field_size)):
                if apple_location not in self.snake and apple_location not in self.apples:
                    self.apples.append(apple_location)
                    break

    def move(self):
        # check what next field would be
        # check if its a wall → lose
        # check if its an apple → grow and move
        # else: it's empty → move
        orientation = self.orientation_switch[self.orientation]
        head = self.snake[-1]
        next_field = (head[0] + orientation[0], head[1] + orientation[1])
        # print(head, next_field)

        if all([0 < pos <= self.field_size for pos in next_field]) and next_field not in self.snake:
            # next_field is not inside a wall, so we can walk
            self.snake.append(next_field)

            # should we delete the last piece?
            if next_field in self.apples:
                self.apples.pop(self.apples.index(next_field))

                for i in range(self.max_apples - len(self.apples)):
                    self.set_apple()
            else:
                self.snake.pop(0)

            return True
        else:
            return False

    def draw_field(self, print_=True) -> Optional[List[List[str]]]:
        entire_field = [[self.theme.border_piece] * (self.field_size + 2)]
        for y in range(1, self.field_size + 1):
            row = [self.theme.border_piece]
            for x in range(1, self.field_size + 1):
                field = (x, y)

                if field in self.apples:
                    row.append(self.theme.apple)
                elif field in self.snake:
                    row.append("🔲")
                else:
                    row.append(self.theme.field_piece)

            row.append(self.theme.border_piece)
            entire_field.append(row)
        entire_field.append([self.theme.border_piece] * (self.field_size + 2))

        for index, (x, y) in enumerate(self.snake):
            if (x, y) == self.snake[-1]:
                entire_field[y][x] = self.theme.head
            elif (x, y) == self.snake[0]:
                entire_field[y][x] = self.theme.tail_end
            else:
                prev_x, prev_y = self.snake[index - 1]
                next_x, next_y = self.snake[index + 1]
                diff_1 = (prev_x - x, prev_y - y)
                diff_2 = (next_x - x, next_y - y)

                orientation_1 = list(self.orientation_switch.values()).index(diff_1)
                orientation_2 = list(self.orientation_switch.values()).index(diff_2)

                string_1 = "urdl"[orientation_1]
                string_2 = "urdl"[orientation_2]

                icon = string_1 + string_2

                entire_field[y][x] = self.theme.tail[icon]

        if print_:
            for row in entire_field:
                for item in row:
                    print(item, end=self.theme.character_separator)
                print()
        else:
            return entire_field

    @property
    def score(self):
        return len(self.snake)

    def add_another_apple(self):
        self.max_apples += 1

    def end_game(self):
        if self.score > self.highscore:
            self.highscore = self.score
            with open("scores.json", "w") as f:
                self.scores[self.score_key] = self.score
                f.write(json.dumps(self.scores))



import curses
import locale
import time
from typing import List

from .game import Game


class TUI:
    KEYBINDINGS = {
        ord("w"): 0,
        ord("a"): 3,
        ord("s"): 2,
        ord("d"): 1,
        curses.KEY_UP: 0,
        curses.KEY_RIGHT: 1,
        curses.KEY_DOWN: 2,
        curses.KEY_LEFT: 3,
    }

    def __init__(self, stdscr, field_size, continous_mode=False):
        self.mode = continous_mode

        locale.setlocale(locale.LC_ALL, '')
        self.stdscr = stdscr
        self.stdscr.border()

        self.stdscr.refresh()

        self.width = curses.COLS
        self.height = curses.LINES

        self.middle = self.middle_y, self.middle_x = (self.height // 2, self.width // 2)

        self.game = self.prepare_game(field_size)

        self.game_pos = self.game_y, self.game_x = (middle_coord - field_coord for middle_coord, field_coord in
                                                    zip(self.middle, (self.game.field_size,) * 2))

        self.game_win = self.stdscr.subwin(self.game_y, self.game_x)
        self.game_win.keypad(1)

        self.msgs = []

        instruction = "Use 'wasd' or your arrow keys to move. Quit using 'q'."
        try:
            self.stdscr.addstr(3, (self.width - len(instruction)) // 2, instruction)
        except curses.error:
            pass
        finally:
            self.stdscr.refresh()

        self.play()

    @staticmethod
    def encode(text: str):
        return text.encode("utf-8", "replace")

    def prepare_game(self, field_size):
        game = Game(field_size, score_key=f"c{field_size}" if self.mode else None)
        game.set_apple()
        return game

    def play(self):
        self.game_win.nodelay(self.mode)

        while True:
            score_text = f"Score: {self.game.score}"
            self.stdscr.addstr(2, (self.width - len(score_text)) - 2, score_text)
            self.stdscr.refresh()

            game_field = self.game.draw_field(print_=False)
            self.game_win.clear()
            for line_count, line in enumerate(game_field):
                self.game_win.addstr(line_count, 0, self.game.theme.character_separator.join(line))
            self.game_win.refresh()

            while (go := self.game_win.getch()) not in list(self.KEYBINDINGS.keys()) + [ord("q")]:
                # wrong chr -> continue if mode == True
                if self.mode:
                    go = [ord(c) for c in "wdsa"][self.game.orientation]
                    break

            if go == ord("q"):
                self.msgs.append("Game aborted.")
                break

            self.game.orientation = self.KEYBINDINGS[go]

            if not self.game.move():
                self.msgs.append("You lost!")
                break

            if self.game.score == self.game.max_score:
                self.msgs.append("You won!")
                break

            self.game.max_apples = max((1, self.game.score // 5))
            if self.mode:
                time.sleep(0.2)
        self.game.end_game()
        self.msgs.append(f"Score: {self.game.score} | Highscore: {self.game.highscore}")



import argparse
from curses import wrapper
import curses

from .game import Game
from .tui import TUI

KEYBINDINGS = {
    "w": 0,
    "a": 3,
    "s": 2,
    "d": 1,
}


def play(game: Game):
    game.set_apple()
    first = True
    while True:
        if not first:
            print("\033[F" * (game.field_size + 4) + "\b")
        else:
            first = False
        game.draw_field()
        lines = 1
        while (go := input("> ").lower()) not in ["w", "a", "s", "d", "exit"]:
            print("\033[F" * (lines + 1) + "\b")
            print("Please type 'w', 'a', 's' or 'd'! To exit the game, type 'exit'")
            lines = 2

        print(("\033[F\b" + " " * 30) * lines)

        if go == "exit":
            print("Game aborted.")
            break

        game.orientation = KEYBINDINGS[go]

        if not game.move():
            print("You lost!")
            break

        if game.score == game.max_score:
            print("You won!")
            break

        game.max_apples = max((1, game.score // 5))
    game.end_game()
    print(f"Score: {game.score} | Highscore: {game.highscore}")


def main(stdscr):
    size = input("Game size: ")
    try:
        size = int(size)
    except ValueError:
        size = 5

    print("\033[F\033[F\b" + " " * 30 + "\b")
    g = Game(size)
    play(g)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Command line snake game')
    parser.add_argument('field_size', type=int, default=5, help='Size of your field')
    parser.add_argument('-c', '--continuous', action='store_true', help='Play this game continuously?')

    args = parser.parse_args()
    print(f"Generating field with size {args.field_size}")

    tui = wrapper(TUI, args.field_size, args.continuous)
    for msg in tui.msgs:
        print(msg)
